import numpy as np
from scipy.sparse import csr_matrix


def dtype_in_datafile(data_df):
    print('Note: Variates beginning with a capital letter is converted into factors.')
    col_names = data_df.columns
    class_vec = []
    for val in col_names:
        if not val[0].isalpha():
            print("The first character of columns names must be alphabet!")
            exit()
        if val[0] == val.capitalize()[0]:
            class_vec.append(val)
            data_df[val] = data_df[val].astype('str')
        else:
            try:
                data_df[val] = data_df[val].astype('float')
            except Exception as e:
                print(e)
                print(val, "may contain string, please check!")
                exit()
    return data_df, class_vec


def check_id_order(id):
    id_order = []
    id_arr = list(id)
    id_order.append(id_arr[0])
    for i in range(1, len(id_arr)):
        if id_arr[i] != id_arr[i - 1]:
            id_order.append(id_arr[i])
    id_in_data = set(id)
    if len(id_in_data) - len(id_order) != 0:
        print('The data is not sored by individual ID!')
        exit()
    return id_order


def recode_fac_lst(fac_lst):
    code_val = 0
    code_dct = {}
    fac_lst_code = []
    for i in range(len(fac_lst)):
        if fac_lst[i] not in code_dct:
            code_val += 1
            code_dct[fac_lst[i]] = str(code_val)
        fac_lst_code.append(code_dct[fac_lst[i]])
    return fac_lst_code, code_dct


def leg_mt(time, tmax, tmin, order):
    time = np.array(time, dtype=float).reshape(max(time.shape), 1)
    tvec = 2 * (time - tmin) / (tmax - tmin) - 1
    pmat = []
    for k in range(order + 1):
        c = int(k / 2)
        j = k
        p = 0
        for r in range(0, c + 1):
            p += np.sqrt((2 * j + 1.0) / 2.0) * pow(0.5, j) * (pow(-1, r) *
                                    np.math.factorial(2 * j - 2 * r) / (np.math.factorial(r) *
                                    np.math.factorial(j - r) * np.math.factorial(j - 2 * r))) * pow(tvec, j - 2 * r)
        p = np.array(p)
        pmat.append(p)
    return pmat


def design_matrix(formula, class_vec, data_df):
    formula_new = formula.replace('+', ' ')
    formula_vec = formula_new.split()
    dmat_lst = [csr_matrix(np.ones((data_df.shape[0], 1)))]
    dmat_len = [1]
    for vari in formula_vec:
        if vari not in data_df.columns:
            print(vari, 'not in the data file!')
            exit()
        if vari in class_vec:
            row = np.arange(data_df.shape[0])
            col = np.array(data_df[vari]) - 1
            val = np.ones(data_df.shape[0])
            dmat_lst.append(csr_matrix((val, (row, col)))[:, 1:])
        else:
            dmat_lst.append(csr_matrix(np.array(data_df[vari]).reshape(data_df.shape[0], 1)))
        dmat_len.append(dmat_lst[-1].shape[1])
    return dmat_lst, dmat_len



import time
from .unbalance_predata import unbalance_predata
from .unbalance_emai import unbalance_emai


def unbalance_varcom(data_file, id, tpoint, trait, gmat_inv_file, tfix=None, fix=None, forder=3, aorder=3, porder=3,
             na_method='omit', fix_const=False, ran_const=True, init=None, max_iter=200, cc_par=1.0e-8, cc_gra=1.0e6,
                     cc_ll=1.0e6, em_weight_step=0.01, prefix_outfile='glta_unbalance_varcom'):
    print("#" * 100)
    print("{:#^100}".format("Prepare the data for unbalanced longitudinal variances estimation"))
    print("#" * 100)
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    y, xmat, zmat, rmat_inv_lst, code_id_dct, id_order, fix_pos_vec = unbalance_predata(data_file, id, tpoint, trait,
                gmat_inv_file, tfix=tfix, fix=fix, forder=forder, aorder=aorder, porder=porder, na_method=na_method)
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    print("&&&Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    
    print("#" * 100)
    print("{:#^100}".format("variances estimation for unbalanced longitudinal data"))
    print("#" * 100)
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    res_df = unbalance_emai(y, xmat, zmat, rmat_inv_lst, fix_const=fix_const, ran_const=ran_const, init=init,
                 max_iter=max_iter, cc_par=cc_par, cc_gra=cc_gra, cc_ll=cc_ll, em_weight_step=em_weight_step)
    res_df['effect_ind'][0] = fix_pos_vec
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    print("&&&Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    
    var_file = prefix_outfile + '.var'
    res_df['variances'].to_csv(var_file, sep=' ', index=False)
    return res_df



import math
import numpy as np
import pandas as pd
from scipy import linalg
from scipy import sparse
from scipy.sparse import csr_matrix, isspmatrix, hstack, vstack


from .unbalance_pre_mat import pre_covar_inv, pre_coef_mat
from .unbalance_iter_mat import unbalance_pre_fd_mat, unbalance_pre_em_mat, unbalance_pre_ai_mat


def unbalance_emai(y, xmat, zmat, rmat_inv_lst, fix_const=False, ran_const=False, init=None, max_iter=30, cc_par=1.0e-8,
                   cc_gra=1.0e-6, cc_ll=0.001, em_weight_step=0.02):
    num_var_diag = 1  # the number of variances (not include non-diag)
    for i in range(len(zmat)):
        num_var_diag += len(zmat[i])
    y_var = np.var(y)/num_var_diag  # for the initial values of variances
    
    num_record = y.shape[0]
    var_com = []
    eff_ind = [[0, xmat.shape[1]]]  # the index for all effects [start end]
    zmat_concat_lst = []  # combined random matrix
    covar_dim = []  # the dim for covariance matrix
    vari = []
    varij = []
    varik = []
    for i in range(len(zmat)):
        temp = [eff_ind[i][-1]]
        zmat_concat_lst.append(hstack(zmat[i]))
        covar_dim.append(len(zmat[i]))
        for j in range(len(zmat[i])):
            temp.append(temp[-1] + zmat[i][j].shape[1])
            for k in range(j + 1):
                vari.append(i+1)
                varij.append(j+1)
                varik.append(k+1)
                if j == k:
                    var_com.append(y_var)
                else:
                    var_com.append(0.0)
        eff_ind.append(temp)
    var_com.append(y_var)
    vari.append(len(zmat) + 1)
    varij.append(1)
    varik.append(1)
    
    if init is None:
        var_com = np.array(var_com)
    else:
        if len(var_com) != len(init):
            print('The length of initial variances should be', len(var_com))
            exit()
        else:
            var_com = np.array(init)
    var_com_update = var_com * 10
    delta = var_com_update - var_com
    
    print('***prepare the MME**')
    zmat_concat = hstack(zmat_concat_lst)  # design matrix for random effects
    wmat = hstack([xmat, zmat_concat])  # merged design matrix
    coef_mat_pure = np.dot(wmat.T, wmat)  # C matrix
    rhs_pure = wmat.T.dot(y)  # right hand
    
    # em weight vector
    if em_weight_step <= 0.0 or em_weight_step > 1.0:
        print('The em weight step should be between 0 (not include) and 1 (include)')
        exit()
    
    iter_count = 0
    
    cc_par_val = 1000.0
    cc_gra_val = 1000.0
    ll_val = 1000.0
    cc_ll_val = 1000.0
    ll_val_pre = 1.0e30
    print("initial variances:", ' '.join(np.array(var_com, dtype=str)))
    
    covar_inv = pre_covar_inv(covar_dim, var_com)
    if covar_inv is None:
        print("Initial variances is not positive define, please check!")
        exit()
    coef_mat_inv, eff, e = pre_coef_mat(y, xmat, eff_ind, zmat_concat, covar_dim, var_com, covar_inv, coef_mat_pure,
                                        rmat_inv_lst, rhs_pure)
    
    while iter_count < max_iter:
        iter_count += 1
        print('***Start the iteration:', str(iter_count) + ' ***')
        # first-order derivative
        fd_mat = unbalance_pre_fd_mat(coef_mat_inv, rmat_inv_lst, covar_inv, eff, eff_ind, e, covar_dim,
                                      zmat_concat_lst, wmat, num_record, var_com)
        # AI matrix
        ai_mat = unbalance_pre_ai_mat(coef_mat_inv, covar_inv, eff, eff_ind, e, covar_dim, zmat_concat_lst, wmat, var_com)
        # EM matrix
        em_mat = unbalance_pre_em_mat(covar_dim, zmat_concat_lst, num_record, var_com)
        
        gamma = -em_weight_step
        while gamma < 1.0:
            gamma = gamma + em_weight_step
            if gamma >= 1.0:
                gamma = 1.0
            wemai_mat = (1 - gamma) * ai_mat + gamma * em_mat
            delta = np.dot(linalg.inv(wemai_mat), fd_mat)
            var_com_update = var_com + delta
            covar_inv = pre_covar_inv(covar_dim, var_com_update)
            if covar_inv is not None:
                print('EM weight value:', str(gamma))
                break
        
        print('Updated variances:', ' '.join(np.array(var_com_update, dtype=str)))
        if covar_inv is None:
            print("Updated variances is not positive define!")
            exit()
        
        coef_mat_inv, eff, e = pre_coef_mat(y, xmat, eff_ind, zmat_concat, covar_dim, var_com_update, covar_inv,
                                            coef_mat_pure, rmat_inv_lst, rhs_pure)
        
        # Convergence criteria
        cc_par_val = np.sum(pow(delta, 2)) / np.sum(pow(var_com_update, 2))
        cc_par_val = np.sqrt(cc_par_val)
        cc_gra_val = np.sqrt(np.sum(pow(fd_mat, 2))) / len(var_com)
        var_com = var_com_update.copy()
        print("Change in parameters, Norm of gradient vector:", str(cc_par_val) + ', ' + str(cc_gra_val))
       
        # Log likelihood function value
        yry = np.divide(np.dot(y.T, y), var_com[-1])
        left = np.divide(wmat.T.dot(y), var_com[-1])
        eff_y = np.dot(coef_mat_inv, left)
        ypy = np.sum(yry - np.dot(left.T, eff_y))
        det_c = -np.linalg.slogdet(coef_mat_inv)[1]
        det_r = num_record * np.log(var_com[-1])
        det_g = 0.0
        for i in range(len(covar_dim)):
            if isspmatrix(rmat_inv_lst[i]):
                det_g += -rmat_inv_lst[i].shape[1] * np.linalg.slogdet(covar_inv[i])[1]
            else:
                det_g += -rmat_inv_lst[i].shape[1] * np.linalg.slogdet(covar_inv[i])[1]
        ll_val = det_r + det_g + det_c + ypy
        print("-2logL: " + str(ll_val))
        cc_ll_val = abs(ll_val_pre - ll_val)
        ll_val_pre = ll_val
        if cc_par_val < cc_par and cc_gra_val < cc_gra and cc_ll_val < cc_ll:
            break
    
    if cc_par_val < cc_par and cc_gra_val < cc_gra and cc_ll_val < cc_ll:
        convergence = True
        print("Variances Converged")
    else:
        convergence = False
        print("Variances not Converged")
    
    # add
    fix_const_val = 0.0
    ran_const_val = 0.0
    if fix_const:
        temp = xmat.T.dot(xmat)
        fix_const_val += -np.linalg.slogdet(temp.toarray())[1]
    if ran_const:
        for i in range(len(covar_dim)):
            if not isspmatrix(rmat_inv_lst[i]):
                ran_const_val += -covar_dim[i] * np.linalg.slogdet(rmat_inv_lst[i])[1]
    ll_val = ll_val + fix_const_val + ran_const_val
    aic_val = 2.0*len(var_com) + ll_val
    print("AIC value: " + str(aic_val))
    bic_val = len(var_com)*math.log(num_record) + ll_val
    print("BIC value: " + str(bic_val))
    var_pd = {'vari': vari,
              "varij": varij,
              "varik": varik,
              "var_val": var_com}
    var_pd = pd.DataFrame(var_pd, columns=['vari', "varij", "varik", "var_val"])
    res = {'variances': var_pd,
           'cc_par': cc_par_val,
           'cc_gra': cc_par_val,
           'convergence': convergence,
           'll': ll_val,
           'effect': eff,
           'effect_ind': eff_ind,
           'coef_inv': coef_mat_inv,
           'AIC': aic_val,
           'BIC': bic_val}
    return res



import numpy as np
from collections import defaultdict


def is_int(num):
    try:
        int(num)
        return True
    except ValueError:
        return False


def is_float(num):
    try:
        float(num)
        return True
    except ValueError:
        return False



def dct_3D():
    # declare the 3D dict
    return defaultdict(dct_31D)


def dct_31D():
    return defaultdict(dct_32D)


def dct_32D():
    return defaultdict()



def dct_2D():
    # declare the 2D dict
    return defaultdict(dct_21D)


def dct_21D():
    return defaultdict()


def dct_1D():
    # declare the 1D dict
    return defaultdict()


def tri_matT(a, b):
    # triple matrix multiplication a x b x a'
    res = np.dot(a, b)
    res = np.dot(res, a.T)
    return res


def tri_mat(a, b, c):
    # triple matrix multiplication a x b x c
    res = np.dot(a, b)
    res = np.dot(res, c)
    return res


def Dtri_matT(a, b):
    # triple matrix multiplication a x D x a'  D is diagonal matrix in row vector form
    res = np.multiply(a, b)
    res = np.dot(res, a.T)
    return res


def Dtri_mat(a, b, c):
    # triple matrix multiplication a x D x c  D is diagonal matrix in row vector form
    res = np.multiply(a, b)
    res = np.dot(res, c)
    return res



import numpy as np
from functools import reduce
import gc
import logging


def em_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com_new = var_com[:]
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df+gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        var_com_new[0] = np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:])) + \
                     np.trace(np.dot(gmat_inv, coef_inv[xmat_df:, xmat_df:]))
        var_com_new[0] = var_com_new[0]/gmat_inv_df
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        var_com_new[1] = np.dot(e_hat, e_hat) + np.trace(np.dot(w_mat, np.dot(coef_inv, w_mat.T)))
        var_com_new[1] = var_com_new[1]/y_df
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta*delta)/np.sum(np.array(var_com_new)*np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com


def pxem_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com_new = var_com[:]
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df+gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        var_com_new[0] = np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:])) + \
                     np.trace(np.dot(gmat_inv, coef_inv[xmat_df:, xmat_df:]))
        var_com_new[0] = var_com_new[0]/gmat_inv_df
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        var_com_new[1] = np.dot(e_hat, e_hat) + np.trace(np.dot(w_mat, np.dot(coef_inv, w_mat.T)))
        var_com_new[1] = var_com_new[1]/y_df
        gamma1 = np.dot(eff[xmat_df:], y - np.dot(xmat, eff[:xmat_df])) - np.trace(np.dot(xmat, coef_inv[:xmat_df, xmat_df:]))
        gamma2 = np.dot(eff[xmat_df:], eff[xmat_df:]) + np.trace(coef_inv[xmat_df:, xmat_df:])
        gamma = gamma1/gamma2
        var_com_new[0] = var_com_new[0]*gamma*gamma
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta*delta)/np.sum(np.array(var_com_new)*np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com


def ai_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com = np.array(var_com)
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df + gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    fd_mat = np.zeros(2)
    # ai_mat = np.zeros((2, 2))
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        fd_mat[0] = gmat_inv_df/var_com[0] - np.trace(np.dot(coef_inv[xmat_df:, xmat_df:], gmat_inv))/(var_com[0]*var_com[0]) - \
            np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:]))/(var_com[0]*var_com[0])
        fd_mat[1] = y_df/var_com[1] - np.trace(np.dot(np.dot(coef_inv, w_mat.T), w_mat))/(var_com[1]*var_com[1]) - \
            np.dot(e_hat, e_hat)/(var_com[1]*var_com[1])
        fd_mat = -0.5*fd_mat
        h_mat1 = np.array(eff[xmat_df:]/var_com[0]).reshape(gmat_inv_df, 1)
        h_mat2 = np.array(e_hat / var_com[1]).reshape(y_df, 1)
        h_mat = np.concatenate((h_mat1, h_mat2), axis=1)
        qrq = np.divide(np.dot(h_mat.T, h_mat), var_com[-1])
        left = np.divide(np.dot(w_mat.T, h_mat), var_com[-1])
        eff_h_mat = np.dot(coef_inv, left)
        ai_mat = np.subtract(qrq, np.dot(left.T, eff_h_mat))
        ai_mat = 0.5 * ai_mat
        ai_mat_inv = np.linalg.inv(ai_mat)
        var_com_new = var_com + np.dot(ai_mat_inv, fd_mat)
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com


def emai_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com = np.array(var_com)
    var_com_new = var_com[:]
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df + gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    fd_mat = np.zeros(2)
    em_mat = np.zeros((2, 2))
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        fd_mat[0] = gmat_inv_df/var_com[0] - np.trace(np.dot(coef_inv[xmat_df:, xmat_df:], gmat_inv))/(var_com[0]*var_com[0]) - \
            np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:]))/(var_com[0]*var_com[0])
        fd_mat[1] = y_df/var_com[1] - np.trace(np.dot(np.dot(coef_inv, w_mat.T), w_mat))/(var_com[1]*var_com[1]) - \
            np.dot(e_hat, e_hat)/(var_com[1]*var_com[1])
        fd_mat = -0.5*fd_mat
        h_mat1 = np.array(eff[xmat_df:]/var_com[0]).reshape(gmat_inv_df, 1)
        h_mat2 = np.array(e_hat / var_com[1]).reshape(y_df, 1)
        h_mat = np.concatenate((h_mat1, h_mat2), axis=1)
        qrq = np.divide(np.dot(h_mat.T, h_mat), var_com[-1])
        left = np.divide(np.dot(w_mat.T, h_mat), var_com[-1])
        eff_h_mat = np.dot(coef_inv, left)
        ai_mat = np.subtract(qrq, np.dot(left.T, eff_h_mat))
        ai_mat = 0.5 * ai_mat
        print(fd_mat, ai_mat)
        em_mat[0, 0] = gmat_inv_df / (var_com[0] * var_com[0])
        em_mat[1, 1] = y_df / (var_com[1] * var_com[1])
        for j in range(0, 51):
            weight = j * 0.1
            wemai_mat = (1 - weight) * ai_mat + weight * em_mat
            delta = np.dot(np.linalg.inv(wemai_mat), fd_mat)
            var_com_new = var_com + delta
            if min(var_com_new) > 0:
                print('EM weight value:', weight)
                break
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com


def pxemai_mme(y, xmat, gmat_inv, init=None, maxiter=100, cc=1.0e-8):
    var_com = [1.0, 1.0]
    if init is not None:
        var_com = init[:]
    var_com = np.array(var_com)
    var_com_new = var_com[:]
    # 准备不含方差组分的系数矩阵
    xmat_df, gmat_inv_df = xmat.shape[1], gmat_inv.shape[0]
    y_df = len(y)
    coef_pre = np.identity(xmat_df + gmat_inv_df)
    coef_pre[:xmat_df, :xmat_df] = np.dot(xmat.T, xmat)
    coef_pre[:xmat_df, xmat_df:] = xmat.T
    coef_pre[xmat_df:, :xmat_df] = xmat
    w_mat = np.concatenate((xmat, np.identity(gmat_inv_df)), axis=1)
    # 开始迭代
    iter = 0
    cc_val = 100.0
    weight = 0.0
    fd_mat = np.zeros(2)
    em_mat = np.zeros((2, 2))
    while iter < maxiter:
        iter += 1
        print('Round: ', iter)
        # 系数矩阵
        coef = coef_pre/var_com[1]
        coef[xmat_df:, xmat_df:] = coef[xmat_df:, xmat_df:] + gmat_inv/var_com[0]
        # 右手项
        rhs_mat = np.dot(w_mat.T, y)/var_com[1]
        coef_inv = np.linalg.inv(coef)
        eff = np.dot(coef_inv, rhs_mat)
        e_hat = y - np.dot(xmat, eff[:xmat_df]) - eff[xmat_df:]
        fd_mat[0] = gmat_inv_df/var_com[0] - np.trace(np.dot(coef_inv[xmat_df:, xmat_df:], gmat_inv))/(var_com[0]*var_com[0]) - \
            np.dot(eff[xmat_df:], np.dot(gmat_inv, eff[xmat_df:]))/(var_com[0]*var_com[0])
        fd_mat[1] = y_df/var_com[1] - np.trace(np.dot(np.dot(coef_inv, w_mat.T), w_mat))/(var_com[1]*var_com[1]) - \
            np.dot(e_hat, e_hat)/(var_com[1]*var_com[1])
        fd_mat = -0.5*fd_mat
        h_mat1 = np.array(eff[xmat_df:]/var_com[0]).reshape(gmat_inv_df, 1)
        h_mat2 = np.array(e_hat / var_com[1]).reshape(y_df, 1)
        h_mat = np.concatenate((h_mat1, h_mat2), axis=1)
        qrq = np.divide(np.dot(h_mat.T, h_mat), var_com[-1])
        left = np.divide(np.dot(w_mat.T, h_mat), var_com[-1])
        eff_h_mat = np.dot(coef_inv, left)
        ai_mat = np.subtract(qrq, np.dot(left.T, eff_h_mat))
        ai_mat = 0.5 * ai_mat
        print(fd_mat, ai_mat)
        em_mat[0, 0] = gmat_inv_df / (var_com[0] * var_com[0])
        em_mat[1, 1] = y_df / (var_com[1] * var_com[1])
        for j in range(0, 51):
            weight = j * 0.1
            wemai_mat = (1 - weight) * ai_mat + weight * em_mat
            delta = np.dot(np.linalg.inv(wemai_mat), fd_mat)
            var_com_new = var_com + delta
            if min(var_com_new) > 0:
                print('EM weight value:', weight)
                break
        if weight > 0.001:
            gamma1 = np.dot(eff[xmat_df:], y - np.dot(xmat, eff[:xmat_df])) - np.trace(
                np.dot(xmat, coef_inv[:xmat_df, xmat_df:]))
            gamma2 = np.dot(eff[xmat_df:], eff[xmat_df:]) + np.trace(coef_inv[xmat_df:, xmat_df:])
            gamma = gamma1 / gamma2
            var_com_new[0] = var_com_new[0] * gamma * gamma
        print('Updated variances:', var_com_new)
        delta = np.array(var_com_new) - np.array(var_com)
        cc_val = np.sum(delta * delta) / np.sum(np.array(var_com_new) * np.array(var_com_new))
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        print('Variances converged.')
    else:
        print('Variances not converged.')
    return var_com


def wemai_multi_gmat(y, xmat, gmat_lst, init=None, maxiter=100, cc=1.0e-8):
    """
    # 单性状多关系矩阵利用EM和AI加权方法估计方差组分
    :param y: 表型向量
    :param xmat: 固定效应设计矩阵
    :param gmat_lst: 关系矩阵列表
    :param init: 初始值
    :param maxiter: 迭代次数
    :param cc: 收敛标准
    :return: 方差组分估计值
    """
    logging.info("预处理")
    var_com = [1.0]*(len(gmat_lst)+1)
    if init is not None:
        var_com = init[:]
    var_com = np.array(var_com)
    var_com_new = var_com[:]
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    logging.info("设置迭代初值")
    iter = 0
    delta = 1000.0
    cc_val = 1000.0
    logging.info("开始迭代")
    while iter < maxiter:
        iter += 1
        logging.info('Round: ' + str(iter))
        # 计算V和其逆
        vmat = np.diag([var_com[-1]]*n)
        for val in range(len(gmat_lst)):
            vmat += gmat_lst[val]*var_com[val]
        vmat = np.linalg.inv(vmat)
        # 计算P矩阵
        vxmat = np.dot(vmat, xmat)
        xvxmat = np.dot(xmat.T, vxmat)
        xvxmat = np.linalg.inv(xvxmat)
        pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
        pmat = vmat - pmat
        pymat = np.dot(pmat, y)
        del vmat, vxmat, xvxmat
        gc.collect()
        # 计算一阶偏导、工具变量
        fd_mat = []
        wv_mat = []
        for val in range(len(gmat_lst)):
            fd_mat_val = -np.trace(np.dot(pmat, gmat_lst[val])) + reduce(np.dot, [pymat.T, gmat_lst[val], pymat])
            fd_mat.append(0.5*np.sum(fd_mat_val))
            wv_mat.append(np.dot(gmat_lst[val], pymat))
        fd_mat_val = -np.trace(pmat) + np.dot(pymat.T, pymat)
        fd_mat.append(0.5 * np.sum(fd_mat_val))
        fd_mat = np.array(fd_mat)
        wv_mat.append(pymat)
        wv_mat = np.concatenate(wv_mat, axis=1)
        del pymat
        gc.collect()
        # AI矩阵
        ai_mat = 0.5*reduce(np.dot, [wv_mat.T, pmat, wv_mat])
        del wv_mat, pmat
        gc.collect()
        em_mat = []
        for val in var_com:
            em_mat.append(n/(val*val))
        em_mat = np.diag(em_mat)
        for j in range(0, 51):
            weight = j * 0.02
            wemai_mat = (1 - weight) * ai_mat + weight * em_mat
            delta = np.dot(np.linalg.inv(wemai_mat), fd_mat)
            var_com_new = var_com + delta
            if min(var_com_new) > 0:
                logging.info('EM weight value: ' + str(weight))
                break
        logging.info('Updated variances: ' + ' '.join(list(np.array(var_com_new, dtype=str))))
        cc_val = np.sum(delta * delta) / np.sum(var_com_new * var_com_new)
        cc_val = np.sqrt(cc_val)
        var_com = var_com_new[:]
        if cc_val < cc:
            break
    if cc_val < cc:
        logging.info('Variances converged.')
    else:
        logging.info('Variances not converged.')
    return var_com



from glta.remma.remma_cpu.remma_add_cpu import remma_add_cpu
from glta.remma.remma_cpu.remma_dom_cpu import remma_dom_cpu

from glta.remma.remma_cpu.remma_epiAA_cpu import remma_epiAA_cpu, remma_epiAA_select_cpu, remma_epiAA_pair_cpu, \
    remma_epiAA_eff_cpu, remma_epiAA_eff_cpu_c

from glta.remma.remma_cpu.remma_epiAD_cpu import remma_epiAD_cpu, remma_epiAD_select_cpu, remma_epiAD_pair_cpu, \
    remma_epiAD_eff_cpu, remma_epiAD_eff_cpu_c

from glta.remma.remma_cpu.remma_epiDD_cpu import remma_epiDD_cpu, remma_epiDD_select_cpu, remma_epiDD_pair_cpu, \
    remma_epiDD_eff_cpu, remma_epiDD_eff_cpu_c



import numpy as np
import pandas as pd
from functools import reduce
import gc
import time
import logging
import sys
from scipy.stats import chi2

from glta.process_plink.process_plink import read_plink, impute_geno


def remma_epiAA_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, p_cut=0.0001, out_file='remma_epiAA_cpu'):
    """
    加加上位检验
    :param y: 表型
    :param xmat: 固定效应设计矩阵
    :param gmat_lst: 基因组关系矩阵列表
    :param var_com: 方差组分
    :param bed_file: plink文件
    :param snp_lst_0: 互作对第一个SNP列表，列表最小值为0，最大值为num_snp-2
    :param p_cut: 依据阈值保留的互作对
    :param out_file: 输出文件
    :return:
    """
    logging.info("计算V矩阵及其逆矩阵")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    del gmat_lst
    gc.collect()
    vmat_inv = np.linalg.inv(vmat)
    logging.info("计算P矩阵")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = np.dot(pmat, y)
    pvpmat = reduce(np.dot, [pmat, vmat, pmat])
    del vmat, vmat_inv, pmat
    gc.collect()
    logging.info("读取SNP文件")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, num_snp)
    snp_mat = snp_mat - 2 * freq
    logging.info('检验')
    if snp_lst_0 is None:
        snp_lst_0 = range(num_snp - 1)
    else:
        if max(snp_lst_0) >= num_snp - 1 or min(snp_lst_0) < 0:
            logging.error('snp_lst_0 is out of range!')
            sys.exit()
    chi2_cut = chi2.isf(p_cut, 1)
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    res_lst = []
    for i in snp_lst_0:
        epi_mat = snp_mat[:, i:(i+1)] * snp_mat[:, (i+1):]
        eff_vec = np.dot(epi_mat.T, pymat)
        var_vec = np.sum(epi_mat * np.dot(pvpmat, epi_mat), axis=0)
        var_vec = var_vec.reshape(len(var_vec), -1)
        chi_vec = eff_vec * eff_vec / var_vec
        res = np.concatenate([np.array([i]*(num_snp-i-1)).reshape(-1, 1), np.arange((i+1), num_snp).reshape(-1, 1),
                              eff_vec,  var_vec, chi_vec], axis=1)
        res_lst.append(res[res[:, -1] > chi2_cut, :])
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    res_lst = np.concatenate(res_lst, axis=0)
    np.savetxt(out_file, res_lst, header='snp_0 snp_1 eff var chi', comments='')
    return res_lst


def remma_epiAA_select_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, snp_lst_1=None, p_cut=1.0, out_file='remma_epiAA_select_cpu'):
    """
    加加上位检验
    :param y: 表型
    :param xmat: 固定效应设计矩阵
    :param gmat_lst: 基因组关系矩阵列表
    :param var_com: 方差组分
    :param bed_file: plink文件
    :param snp_lst_0: 互作对第一个SNP列表,列表最小值为0，最大值为num_snp-1
    :param snp_lst_1: 互作对第一个SNP列表,列表最小值为0，最大值为num_snp-1
    :param p_cut: 依据阈值保留的互作对
    :param out_file: 输出文件
    :return:
    """
    logging.info("计算V矩阵及其逆矩阵")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    del gmat_lst
    gc.collect()
    vmat_inv = np.linalg.inv(vmat)
    logging.info("计算P矩阵")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = np.dot(pmat, y)
    pvpmat = reduce(np.dot, [pmat, vmat, pmat])
    del vmat, vmat_inv, pmat
    gc.collect()
    logging.info("读取SNP文件")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, num_snp)
    snp_mat = snp_mat - 2 * freq
    logging.info("开始检验")
    if snp_lst_0 is None:
        snp_lst_0 = range(num_snp)
    else:
        if max(snp_lst_0) > num_snp - 1 or min(snp_lst_0) < 0:
            logging.error('snp_lst_0 is out of range!')
            sys.exit()
    if snp_lst_1 is None:
        snp_lst_1 = range(num_snp)
    else:
        if max(snp_lst_1) > num_snp - 1 or min(snp_lst_1) < 0:
            logging.error('snp_lst_1 is out of range!')
            sys.exit()
    snp_lst_0 = list(snp_lst_0)
    snp_lst_1 = list(snp_lst_1)
    chi2_cut = chi2.isf(p_cut, 1)
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    res_lst = []
    for i in snp_lst_0:
        snp_lst_11 = snp_lst_1[:]
        try:
            snp_lst_11.remove(i)
        except Exception as e:
            del e
        epi_mat = snp_mat[:, i:(i + 1)] * snp_mat[:, snp_lst_11]
        eff_vec = np.dot(epi_mat.T, pymat)
        var_vec = np.sum(epi_mat * np.dot(pvpmat, epi_mat), axis=0)
        var_vec = var_vec.reshape(-1, 1)
        chi_vec = eff_vec * eff_vec / var_vec
        res = np.concatenate([np.array([i] * len(snp_lst_11)).reshape(-1, 1), np.array(snp_lst_11).reshape(-1, 1),
                              eff_vec, var_vec, chi_vec], axis=1)
        res_lst.append(res[res[:, -1] > chi2_cut, :])
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    res_lst = np.concatenate(res_lst, axis=0)
    np.savetxt(out_file, res_lst, header='snp_0 snp_1 eff var chi', comments='')
    return res_lst


def remma_epiAA_pair_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_pair_file, max_test_pair=50000, p_cut=1.0e-4, out_file='remma_epiAA_pair_cpu'):
    """
    加加上位检验
    :param y: 表型
    :param xmat: 固定效应设计矩阵
    :param gmat_lst: 基因组关系矩阵列表
    :param var_com: 方差组分
    :param bed_file: plink文件
    :param snp_pair_file: SNP互作对文件，取前两列，最小值为0，最大值为num_snp-1
    :param max_test_pair: 分批次检验，每次检验的最大互作对数
    :param p_cut: 阈值
    :param out_file: 输出文件
    :return:
    """
    logging.info("计算V矩阵及其逆矩阵")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    del gmat_lst
    gc.collect()
    vmat_inv = np.linalg.inv(vmat)
    logging.info("计算P矩阵")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = np.dot(pmat, y)
    pvpmat = reduce(np.dot, [pmat, vmat, pmat])
    del vmat, vmat_inv, pmat
    gc.collect()
    logging.info("读取SNP文件")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, -1)
    snp_mat = snp_mat - 2 * freq
    logging.info("开始检验")
    np.savetxt(out_file, ['snp_0 snp_1 eff var chi p'], fmt='%s')
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    ipart = -1
    while True:
        ipart += 1
        skiprows = 1 + ipart*max_test_pair
        try:
            snp_pair = pd.read_csv(snp_pair_file, header=None, sep='\s+', skiprows=skiprows, nrows=max_test_pair)
        except Exception as e:
            logging.info(e)
            break
        snp_pair = np.array(snp_pair.iloc[:, 0:2], dtype=np.int)
        if np.max(snp_pair) > num_snp - 1 or np.min(snp_pair) < 0:
            logging.error('snp_pair is out of range!')
            sys.exit()
        epi_mat = snp_mat[:, snp_pair[:, 0]] * snp_mat[:, snp_pair[:, 1]]
        eff_vec = np.dot(epi_mat.T, pymat)
        var_vec = np.sum(epi_mat * np.dot(pvpmat, epi_mat), axis=0)
        var_vec = var_vec.reshape(-1, 1)
        chi_vec = eff_vec * eff_vec / var_vec
        p_vec = chi2.sf(chi_vec, 1)
        res = pd.DataFrame({0: snp_pair[:, 0], 1: snp_pair[:, 1], 2: eff_vec[:, -1], 3: var_vec[:, -1], 4: chi_vec[:, -1], 5: p_vec[:, -1]})
        res = res[res[5] < p_cut]
        res.to_csv(out_file, sep=' ', header=False, index=False, mode='a')
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info(
        "Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    return 0


def remma_epiAA_eff_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu'):
    """
    加加上位检验
    :param y: 表型
    :param xmat: 固定效应设计矩阵
    :param gmat_lst: 基因组关系矩阵列表
    :param var_com: 方差组分
    :param bed_file: plink文件
    :param snp_lst_0: 互作对第一个SNP列表，最小值为0，最大值为num_snp-2
    :param eff_cut: 依据阈值保留的互作对
    :param out_file: 输出文件
    :return:
    """
    logging.info("计算V矩阵及其逆矩阵")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    del gmat_lst
    gc.collect()
    vmat_inv = np.linalg.inv(vmat)
    logging.info("计算P矩阵")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = np.dot(pmat, y)
    del vmat, vmat_inv, pmat
    gc.collect()
    logging.info("读取SNP文件")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, -1)
    snp_mat = snp_mat - 2 * freq
    logging.info('检验')
    if snp_lst_0 is None:
        snp_lst_0 = range(num_snp - 1)
    else:
        if max(snp_lst_0) >= num_snp - 1 or min(snp_lst_0) < 0:
            logging.error('snp_lst_0 is out of range!')
            sys.exit()
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    res_lst = []
    for i in snp_lst_0:
        epi_mat = snp_mat[:, i:(i+1)] * snp_mat[:, (i+1):]
        eff_vec = np.dot(epi_mat.T, pymat)
        res = np.concatenate([np.array([i]*(num_snp-i-1)).reshape(-1, 1), np.arange((i+1), num_snp).reshape(-1, 1), eff_vec], axis=1)
        res_lst.append(res[np.abs(res[:, -1]) > eff_cut, :])
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    res_lst = np.concatenate(res_lst, axis=0)
    np.savetxt(out_file, res_lst, header='snp_0 snp_1 eff', comments='')
    return res_lst


from _cremma_epi_eff_cpu import ffi, lib


def remma_epiAA_eff_cpu_c(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu_c'):
    """
    加加上位检验
    :param y: 表型
    :param xmat: 固定效应设计矩阵
    :param gmat_lst: 基因组关系矩阵列表
    :param var_com: 方差组分
    :param bed_file: plink文件
    :param snp_lst_0: 互作对第一个SNP列表，最大值为num_snp-2，最小值为0
    :param eff_cut: 依据阈值保留的互作对
    :param out_file: 输出文件
    :return:
    """
    logging.info("计算V矩阵及其逆矩阵")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    del gmat_lst
    gc.collect()
    vmat_inv = np.linalg.inv(vmat)
    logging.info("计算P矩阵")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = np.dot(pmat, y)
    del vmat, vmat_inv, pmat
    gc.collect()
    num_snp = pd.read_csv(bed_file+'.bim', header=None).shape[0]
    num_id = pd.read_csv(bed_file+'.fam', header=None).shape[0]
    if snp_lst_0 is None:
        snp_lst_0 = range(num_snp - 1)
    else:
        if max(snp_lst_0) >= num_snp - 1 or min(snp_lst_0) < 0:
            logging.error('snp_lst_0 is out of range!')
            sys.exit()
    logging.info("python变量转化到C")
    pbed_file = ffi.new("char[]", bed_file.encode('ascii'))
    pnum_id = ffi.cast("long long", num_id)
    pnum_snp = ffi.cast("long long", num_snp)
    snp_lst_0 = np.array(list(snp_lst_0), dtype=np.longlong)
    psnp_lst_0 = ffi.cast("long long *", snp_lst_0.ctypes.data)
    # psnp_lst_0 = ffi.cast("long long *", ffi.from_buffer(snp_lst_0))
    plen_snp_lst_0 = ffi.cast("long long", len(snp_lst_0))
    ppymat = ffi.cast("double *", pymat.ctypes.data)
    peff_cut = ffi.cast("double", eff_cut)
    pout_file = ffi.new("char[]", out_file.encode('ascii'))
    logging.info('检验')
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    lib.remma_epiAA_eff_cpu(pbed_file, pnum_id, pnum_snp, psnp_lst_0, plen_snp_lst_0, ppymat, peff_cut, pout_file)
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    return 0



import numpy as np
import pandas as pd
from functools import reduce
import gc
import time
import logging
from scipy.stats import chi2

from glta.process_plink.process_plink import read_plink, impute_geno


def remma_epiAA_eff_gpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, max_test_pair=50000, eff_cut=-999.0, out_file='remma_epiAA_eff_gpu'):
    """
    加加上位检验，GPU加速
    :param y: 表型
    :param xmat: 固定效应设计矩阵
    :param gmat_lst: 基因组关系矩阵列表
    :param var_com: 方差组分
    :param bed_file: plink文件
    :param snp_lst_0: 互作对第一个SNP列表
    :param max_test_pair: 最大检验互作对数
    :param eff_cut: 依据阈值保留的互作对
    :param out_file: 输出文件
    :return:
    """
    try:
        import cupy as cp
    except Exception as e:
        logging.error(e)
        return e
    logging.info("计算V矩阵及其逆矩阵")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    vmat_inv = np.linalg.inv(vmat)
    logging.info("计算P矩阵")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = np.dot(pmat, y)
    del vmat, vmat_inv, pmat
    gc.collect()
    logging.info("读取SNP文件")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, num_snp)
    snp_mat = snp_mat - 2 * freq
    logging.info('检验')
    if snp_lst_0 is None:
        snp_lst_0 = range(num_snp - 1)
    else:
        if max(snp_lst_0) >= num_snp - 1 or min(snp_lst_0) < 0:
            logging.error('snp_lst_0 is out of range!')
            sys.exit()
    snp_mat0 = cp.array(snp_mat[:, snp_lst_0])
    pymat = cp.array(pymat)
    clock_t0 = time.perf_counter()
    cpu_t0 = time.process_time()
    res_lst = []
    start, end = 0, 0
    while True:
        start = end
        if start >= num_snp:
            break
        end = start + max_test_pair
        if end >= num_snp:
            end = num_snp
        snp_mat1 = cp.array(snp_mat[:, start:end])
        num_snp1 = snp_mat1.shape[1]
        for i in range(len(snp_lst_0)):
            if end >= i+2 and start <= i:
                epi_mat = snp_mat0[:, i:(i + 1)] * snp_mat1[:, (i+1):]
                eff_vec = cp.dot(epi_mat.T, pymat)
                res = cp.concatenate([cp.array([snp_lst_0[i]] * (snp_mat1.shape[1] - i - 1)).reshape(-1, 1),
                                      cp.arange(i+1, snp_mat1.shape[1]).reshape(-1, 1), eff_vec], axis=1)
                res_lst.append(res[cp.absolute(res[:, -1]) > eff_cut, :])
            elif start > i:
                epi_mat = snp_mat0[:, i:(i + 1)] * snp_mat1
                eff_vec = cp.dot(epi_mat.T, pymat)
                res = cp.concatenate([cp.array([snp_lst_0[i]] * num_snp1).reshape(-1, 1),
                                  cp.arange(start, end).reshape(-1, 1), eff_vec], axis=1)
                res_lst.append(res[cp.absolute(res[:, -1]) > eff_cut, :])
            else:
                continue
    clock_t1 = time.perf_counter()
    cpu_t1 = time.process_time()
    logging.info("Running time: Clock time, {:.5f} sec; CPU time, {:.5f} sec.".format(clock_t1 - clock_t0, cpu_t1 - cpu_t0))
    res_lst = cp.asnumpy(cp.concatenate(res_lst, axis=0))
    np.savetxt(out_file, res_lst, header='snp_0 snp_1 eff', comments='')
    return res_lst



import pandas as pd
import numpy as np
from scipy.stats import chi2
import logging
from glta.gmat import agmat, dgmat_as
from glta.uvlmm import wemai_multi_gmat

from glta.remma.remma_cpu import remma_add_cpu
from glta.remma.remma_cpu import remma_dom_cpu

from glta.remma.remma_cpu import remma_epiAA_cpu, remma_epiAA_select_cpu, remma_epiAA_pair_cpu, \
    remma_epiAA_eff_cpu, remma_epiAA_eff_cpu_c

from glta.remma.remma_cpu import remma_epiAD_cpu, remma_epiAD_select_cpu, remma_epiAD_pair_cpu, \
    remma_epiAD_eff_cpu, remma_epiAD_eff_cpu_c

from glta.remma.remma_cpu import remma_epiDD_cpu, remma_epiDD_select_cpu, remma_epiDD_pair_cpu, \
    remma_epiDD_eff_cpu, remma_epiDD_eff_cpu_c




logging.basicConfig(level=logging.INFO)



bed_file = '../data/plink'
df = pd.read_csv("../data/pheno", header=None, sep='\s+')

y = np.array(df.iloc[:, -1:])
xmat = np.array(df.iloc[:, 2:-1])



agmat_lst = agmat(bed_file, inv=True, small_val=0.001, out_fmt='mat')
dgmat_lst = dgmat_as(bed_file, inv=True, small_val=0.001, out_fmt='mat')


gmat_lst = [agmat_lst[0], dgmat_lst[0], agmat_lst[0]*agmat_lst[0], agmat_lst[0]*dgmat_lst[0], dgmat_lst[0]*dgmat_lst[0]]

var_com = wemai_multi_gmat(y, xmat, gmat_lst, init=None, maxiter=200, cc=1.0e-8)


res_add = remma_add_cpu(y, xmat, gmat_lst, var_com, bed_file, out_file='remma_add_cpu')
print(np.median(res_add['p_val']))

res_dom = remma_dom_cpu(y, xmat, gmat_lst, var_com, bed_file, out_file='remma_dom_cpu')
print(np.median(res_dom['p_val']))

###########加加互作
res_epiAA_cpu = remma_epiAA_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, p_cut=1, out_file='remma_epiAA_cpu')
p_vec = chi2.sf(res_epiAA_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiAA_select_cpu = remma_epiAA_select_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, snp_lst_1=None, p_cut=1.0, out_file='remma_epiAA_select_cpu')
p_vec = chi2.sf(res_epiAA_select_cpu[:, -1], 1)
print(np.median(p_vec))


snp1 = np.arange(1, 1407).reshape(-1, 1)
snp2 = np.array([0]*1406).reshape(-1, 1)
snp_pair = np.concatenate([snp2, snp1], axis=1)
res_epiAA_pair_cpu = remma_epiAA_pair_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_pair, out_file='remma_epiAA_pair_cpu')
p_vec = chi2.sf(res_epiAA_pair_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiAA_eff_cpu = remma_epiAA_eff_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu')

res_epiAA_eff_cpu_c = remma_epiAA_eff_cpu_c(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu_c')


###########加显互作
res_epiAD_cpu = remma_epiAD_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, p_cut=1, out_file='remma_epiAD_cpu')
p_vec = chi2.sf(res_epiAD_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiAD_select_cpu = remma_epiAD_select_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, snp_lst_1=None, p_cut=1.0, out_file='remma_epiAD_select_cpu')
p_vec = chi2.sf(res_epiAD_select_cpu[:, -1], 1)
print(np.median(p_vec))


snp1 = np.arange(1, 1407).reshape(-1, 1)
snp2 = np.array([0]*1406).reshape(-1, 1)
snp_pair = np.concatenate([snp2, snp1], axis=1)
res_epiAD_pair_cpu = remma_epiAD_pair_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_pair, out_file='remma_epiAD_pair_cpu')
p_vec = chi2.sf(res_epiAD_pair_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiAD_eff_cpu = remma_epiAD_eff_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAD_eff_cpu')

res_epiAD_eff_cpu_c = remma_epiAD_eff_cpu_c(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAD_eff_cpu_c')

###########显显互作
res_epiDD_cpu = remma_epiDD_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, p_cut=1, out_file='remma_epiDD_cpu')
p_vec = chi2.sf(res_epiDD_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiDD_select_cpu = remma_epiDD_select_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, snp_lst_1=None, p_cut=1.0, out_file='remma_epiDD_select_cpu')
p_vec = chi2.sf(res_epiDD_select_cpu[:, -1], 1)
print(np.median(p_vec))


snp1 = np.arange(1, 1407).reshape(-1, 1)
snp2 = np.array([0]*1406).reshape(-1, 1)
snp_pair = np.concatenate([snp2, snp1], axis=1)
res_epiDD_pair_cpu = remma_epiDD_pair_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_pair, out_file='remma_epiDD_pair_cpu')
p_vec = chi2.sf(res_epiDD_pair_cpu[:, -1], 1)
print(np.median(p_vec))

res_epiDD_eff_cpu = remma_epiDD_eff_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiDD_eff_cpu')

res_epiDD_eff_cpu_c = remma_epiDD_eff_cpu_c(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiDD_eff_cpu_c')







var_com = [0.06289206, 0.07641075, 0.08121168]
# res_epiAA_eff_cpu = remma_epiAA_eff_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu')

var_com = [0.06289206, 0.07641075, 0.08121168]
res_epiAA_eff_cpu_c = remma_epiAA_eff_cpu_c(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, eff_cut=-999.0, out_file='remma_epiAA_eff_cpu_c')

'''
res_epiAA_select_cpu = remma_epiAA_select_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_lst_0=None, snp_lst_1=None, p_cut=1.0, out_file='remma_epiAA_select_cpu')


snp1 = np.arange(1407).reshape(-1, 1)
snp2 = np.array([0]*1407).reshape(-1, 1)
snp_pair = np.concatenate([snp2, snp1], axis=1)
res_epiAA_pair_cpu = remma_epiAA_pair_cpu(y, xmat, gmat_lst, var_com, bed_file, snp_pair, out_file='remma_epiAA_pair_cpu')


from glta.remma import remma_random_pair

num_snp = 1000
remma_random_pair(num_snp, num_pair=100000, out_file='random_pair')


'''


import numpy as np
from scipy import linalg
from scipy import sparse
from scipy.sparse import isspmatrix


def unbalance_pre_fd_mat(coef_mat_inv, kin, covar_inv, eff, eff_ind, e, covar_dim, zmat_concat_lst, wmat, num_record,
                         var_com):
    fd_mat = []
    for i in range(len(covar_dim)):
        qnum = zmat_concat_lst[i].shape[1] / covar_dim[i]
        tmat = np.array([0.0] * covar_dim[i] * covar_dim[i]).reshape(covar_dim[i], covar_dim[i])
        eff_lst = []
        if not isspmatrix(kin[i]):
            for j in range(covar_dim[i]):
                for k in range(j + 1):
                    tmat[j, k] = np.sum(np.multiply(kin[i], coef_mat_inv[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1],
                                                            eff_ind[i + 1][k]:eff_ind[i + 1][k + 1]]))
                    tmat[k, j] = tmat[j, k]
                eff_lst.append(eff[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1], :])
            eff_mat = np.dot(np.concatenate(eff_lst, axis=1), covar_inv[i])
            temp = qnum * covar_inv[i] - np.dot(np.dot(covar_inv[i], tmat), covar_inv[i]) \
                   - np.dot(np.dot(eff_mat.T, kin[i]), eff_mat)
        else:
            for j in range(covar_dim[i]):
                for k in range(j + 1):
                    tmat[j, k] = np.sum(kin[i].multiply(coef_mat_inv[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1],
                                                        eff_ind[i + 1][k]:eff_ind[i + 1][k + 1]]))
                    tmat[k, j] = tmat[j, k]
                eff_lst.append(eff[eff_ind[i + 1][j]:eff_ind[i + 1][j + 1], :])
            eff_mat = np.dot(np.concatenate(eff_lst, axis=1), covar_inv[i])
            temp = qnum * covar_inv[i] - np.dot(np.dot(covar_inv[i], tmat), covar_inv[i]) \
                   - np.dot(eff_mat.T, kin[i].dot(eff_mat))
        
        temp[np.tril_indices(covar_dim[i], k=-1)] = 2.0 * temp[np.tril_indices(covar_dim[i], k=-1)]
        fd_mat.extend(list(-0.5 * temp[np.tril_indices(covar_dim[i], k=0)]))
    
    temp = -0.5 * (num_record / var_com[-1] - np.sum((wmat.T.dot(wmat)).multiply(coef_mat_inv)) \
                   / (var_com[-1] * var_com[-1]) - np.sum(np.dot(e.T, e) / (var_com[-1] * var_com[-1])))
    
    fd_mat.append(temp)
    # print "FD matrix:", fd_mat
    fd_mat = np.array(fd_mat)
    return fd_mat


def unbalance_pre_ai_mat(coef_mat_inv, covar_inv, eff, eff_ind, e, covar_dim, zmat_concat_lst, wmat, var_com):
    wv = []   # working variable
    for i in range(len(covar_dim)):
        dial = sparse.eye(zmat_concat_lst[i].shape[1] / covar_dim[i], dtype=np.float64)
        for j in range(covar_dim[i]):
            for k in range(j + 1):
                var_fd = np.zeros((covar_dim[i], covar_dim[i]))  # Notice
                var_fd[j, k] = var_fd[k, j] = 1.0
                temp = sparse.kron(np.dot(var_fd, covar_inv[i]), dial)
                temp = temp.dot(eff[eff_ind[i + 1][0]:eff_ind[i + 1][-1], :])
                temp = zmat_concat_lst[i].dot(temp)
                wv.append(temp)
    wv.append(e / var_com[-1])
    qmat = np.concatenate(wv, axis=1)
    
    # AI matrix
    qrq = np.divide(np.dot(qmat.T, qmat), var_com[-1])
    left = np.divide(wmat.T.dot(qmat), var_com[-1])
    eff_qmat = np.dot(coef_mat_inv, left)
    ai_mat = np.subtract(qrq, np.dot(left.T, eff_qmat))
    ai_mat = 0.5 * ai_mat
    try:
        linalg.cholesky(ai_mat)
    except Exception as e:
        print(e)
        ai_mat = False
    return ai_mat


def unbalance_pre_em_mat(covar_dim, zmat_concat_lst, num_record, var_com):
    num_var_com = len(var_com)
    em_mat = np.array([0.0] * num_var_com * num_var_com).reshape(num_var_com, num_var_com)
    b = 0
    cov_mat = []
    for i in range(len(covar_dim)):
        ind = np.tril_indices(covar_dim[i])
        temp = np.array([0.0] * len(ind[0]) * len(ind[0])).reshape(len(ind[0]), len(ind[0]))
        q = zmat_concat_lst[i].shape[1] / covar_dim[i]
        a = b
        b = b + len(ind[0])
        covar = np.zeros((covar_dim[i], covar_dim[i]))
        covar[ind] = var_com[a:b]
        covar = np.add(covar, np.tril(covar, -1).T)
        cov_mat.append(covar)
        for j in range(len(ind[0])):
            for k in range(j + 1):
                temp[j, k] = (cov_mat[i][ind[0][j], ind[0][k]] * cov_mat[i][ind[1][j], ind[1][k]] + \
                              cov_mat[i][ind[0][j], ind[1][k]] * cov_mat[i][ind[1][j], ind[0][k]]) / (2.0 * q)
        em_mat[a:b, a:b] = temp.copy()
    em_mat[-1, -1] = (var_com[-1] * var_com[-1]) / num_record
    em_mat = 2.0 * em_mat
    
    em_mat += np.tril(em_mat, k=-1).T
    return linalg.inv(em_mat)




### gma.cal_kin
from gma import cal_kin
bed_file = './tests/plink'
kin_lst = cal_kin(bed_file)


### gma.balance_varcom
import numpy as np
import pandas as pd
from gma import cal_kin
from gma import balance_varcom
bed_file = './tests/plink'
kin_lst = cal_kin(bed_file)
data_file = './tests/phe.balance.txt'
id = 'ID'
tpoint = np.array(range(16)) + 1.0
trait = range(2, 18)
kin_file = './tests/plink.grm'
prefix_outfile = './tests/gma_balance_varcom'
res_var = balance_varcom(data_file, id, tpoint, trait, kin_file, prefix_outfile=prefix_outfile)
print res_var
# If the variances do not converge, we can use the previous variances as initial values.
init = np.array(res_var['variances']['var_val'])
# or
# var_com = pd.read_csv('./tests/gma_balance_varcom.var', header=0, sep='\s+')
# init = np.array(var_com['var_val'])
res_var = balance_varcom(data_file, id, tpoint, trait, kin_file, init=init, prefix_outfile=prefix_outfile)


### gma.balance_longwas_fixreg
import numpy as np
import pandas as pd
from gma import cal_kin
from gma import balance_varcom
from gma import balance_longwas_fixreg
bed_file = './tests/plink'
kin_lst = cal_kin(bed_file)
data_file = './tests/phe.balance.txt'
id = 'ID'
tpoint = np.array(range(16)) + 1.0
trait = range(2, 18)
kin_file = './tests/plink.grm'
prefix_outfile = './tests/gma_balance_varcom'
res_var = balance_varcom(data_file, id, tpoint, trait, kin_file, prefix_outfile=prefix_outfile)
var_com = res_var['variances']
# or
# var_com = pd.read_csv('./tests/gma_balance_varcom.var', header=0, sep='\s+')
snp_lst = range(1, 10)
# snp_lst = [2, 3, 5, 10]
prefix_outfile = './tests/gma_balance_longwas_fixreg'
res_lst = balance_longwas_fixreg(data_file, id, tpoint, trait, kin_file, bed_file, var_com, snp_lst=snp_lst,
                                 prefix_outfile=prefix_outfile)

### gma.balance_longwas_lt
import numpy as np
import pandas as pd
from gma import cal_kin
from gma import balance_varcom
from gma import balance_longwas_lt
bed_file = './tests/plink'
kin_lst = cal_kin(bed_file)
data_file = './tests/phe.balance.txt'
id = 'ID'
tpoint = np.array(range(16)) + 1.0
trait = range(2, 18)
kin_file = './tests/plink.grm'
prefix_outfile = './tests/gma_balance_varcom'
res_var = balance_varcom(data_file, id, tpoint, trait, kin_file, prefix_outfile=prefix_outfile)
var_com = res_var['variances']
# or
# var_com = pd.read_csv('./tests/gma_balance_varcom.var', header=0, sep='\s+')
snp_lst = range(1, 10)
# snp_lst = [2, 3, 5, 10]
prefix_outfile = './tests/gma_balance_longwas_lt'
res_lst = balance_longwas_lt(data_file, id, tpoint, trait, kin_file, bed_file, var_com, snp_lst=snp_lst,
                                 prefix_outfile=prefix_outfile)


### gma.balance_longwas_lt + gma.balance_longwas_fixreg
import numpy as np
import pandas as pd
from gma import cal_kin
from gma import balance_varcom
from gma import balance_longwas_lt
bed_file = './tests/plink'
kin_lst = cal_kin(bed_file)
data_file = './tests/phe.balance.txt'
id = 'ID'
tpoint = np.array(range(16)) + 1.0
trait = range(2, 18)
kin_file = './tests/plink.grm'
prefix_outfile = './tests/gma_balance_varcom'
res_var = balance_varcom(data_file, id, tpoint, trait, kin_file, prefix_outfile=prefix_outfile)
var_com = res_var['variances']
# or
# var_com = pd.read_csv('./tests/gma_balance_varcom.var', header=0, sep='\s+')
prefix_outfile = './tests/gma_balance_longwas_lt'
res_lst = balance_longwas_lt(data_file, id, tpoint, trait, kin_file, bed_file, var_com, snp_lst=None,
                                 prefix_outfile=prefix_outfile)
res_lst_sel = res_lst[res_lst['p_val'] < 0.01]
snp_lst = np.array(res_lst_sel['order'])
prefix_outfile = './tests/gma_balance_longwas_fixreg_sel'
res_lst_sel = balance_longwas_fixreg(data_file, id, tpoint, trait, kin_file, bed_file, var_com, snp_lst=snp_lst,
                                 prefix_outfile=prefix_outfile)





import numpy as np
import pandas as pd
from scipy import sparse
from scipy.sparse import csr_matrix, hstack
from glta.longwas.common import dtype_in_datafile, check_id_order, recode_fac_lst, leg_mt, design_matrix


def unbalance_predata(data_file, id, tpoint, trait, gmat_inv_file, tfix=None, fix=None, forder=3, aorder=3,
                      porder=3, na_method='omit'):
    data_df = pd.DataFrame()
    try:
        data_df = pd.read_csv(data_file, sep='\s+', header=0)
    except Exception as e:
        print(e)
        print("Fail to open the data file.")
        exit()
    
    print('How to process missing data:', na_method)
    if na_method == 'omit':
        data_df = data_df.dropna()
    elif na_method == 'include':
        data_df = data_df.fillna(method='ffill')
        data_df = data_df.fillna(method='bfill')
    else:
        print('{} method does not exist'.format(na_method))
        exit()
    
    col_names = data_df.columns
    print('The column names of data file:', ' '.join(list(col_names)))
    data_df, class_vec = dtype_in_datafile(data_df)
    
    print('Individual column:', id)
    if id not in col_names:
        print(id, 'is not in the data file, please check!')
        exit()
    if id not in class_vec:
        print('The initial letter of', id, 'should be capital!')
        exit()
    id_order = check_id_order(data_df[id])
    
    print('Time points column:', tpoint)
    if tpoint not in col_names:
        print(tpoint, 'is not in the data file, please check!')
        exit()
    if tpoint in class_vec:
        print('The initial letter of', tpoint, 'should be lowercase')
        exit()
    
    print('Trait column: ', trait)
    if trait not in col_names:
        print(trait, 'is not in the data file, please check!')
        exit()
    if trait in class_vec:
        print('The initial letter of', trait, 'should be lowercase')
        exit()
    y = np.array(data_df[trait]).reshape(data_df.shape[0], 1)
    
    print('Recode factor variables with integer in the data file:', ' '.join(list(class_vec)))
    code_fac_dct = {}
    for val in class_vec:
        fac_lst_code, code_dct = recode_fac_lst(list(data_df[val]))
        code_fac_dct[val] = code_dct
        data_df[val] = np.array(fac_lst_code, dtype=np.int64)
    
    print('Time dependent fix effect:', str(tfix))
    print('Legendre order is:', forder)
    leg_fix = leg_mt(data_df[tpoint], max(data_df[tpoint]), min(data_df[tpoint]), forder)
    fix_pos_vec = [0]
    if tfix is None:
        xmat_t = np.concatenate(leg_fix, axis=1)
        xmat_t = csr_matrix(xmat_t)
        tfix_len = np.array([forder + 1])
    else:
        tfix_mat, tfix_len = design_matrix(tfix, class_vec, data_df)
        tfix_len = np.array(tfix_len) * (forder + 1)
        xmat_t = []
        for veci in range(len(tfix_mat)):
            for vecj in range(len(leg_fix)):
                xmat_t.append(tfix_mat[veci].multiply(leg_fix[vecj]))
        xmat_t = hstack(xmat_t)
    for i in list(tfix_len):
        fix_pos_vec.append(fix_pos_vec[-1] + i)
    
    print('Time independent fix effect:', str(fix))
    if fix is None:
        xmat_nt = None
    else:
        xmat_nt, fix_len = design_matrix(fix, class_vec, data_df)
        xmat_nt = hstack(xmat_nt[1:])
        for i in list(fix_len[1:]):
            fix_pos_vec.append(fix_pos_vec[-1] + i)
    xmat = hstack([xmat_t, xmat_nt])
    
    print('Read the inversion of kinship matrix')
    row = []
    col = []
    kin_inv = []
    id_in_kin = {}
    code_val = max(np.array(list(code_fac_dct[id].values())).astype(np.int))
    with open(gmat_inv_file, 'r') as fin:
        for line in fin:
            arr = line.split()
            id_in_kin[arr[0]] = 1
            id_in_kin[arr[1]] = 1
            if arr[0] not in code_fac_dct[id]:
                code_val += 1
                code_fac_dct[id][arr[0]] = str(code_val)
            if arr[1] not in code_fac_dct[id]:
                code_val += 1
                code_fac_dct[id][arr[1]] = str(code_val)
            row.append(int(code_fac_dct[id][arr[0]]))
            col.append(int(code_fac_dct[id][arr[1]]))
            kin_inv.append(float(arr[2]))
    kin_inv = csr_matrix((np.array(kin_inv), (np.array(row) - 1, np.array(col) - 1))).toarray()
    kin_inv = np.add(kin_inv, kin_inv.T)  # maybe not lower triangular matrix
    np.fill_diagonal(kin_inv, 0.5 * np.diag(kin_inv))
    id_in_kin = set(id_in_kin.keys())
    id_in_data = set(id_order)
    id_not_in_kin = list(id_in_data - id_in_kin)
    if len(id_not_in_kin) != 0:
        print('The ID:', id_not_in_kin, 'in the data file is not in the kinship file!')
        exit()
    
    print('Build the design matrix for random effect.')
    print('Legendre order for additive effects:', str(aorder))
    leg_add = leg_mt(data_df[tpoint], max(data_df[tpoint]), min(data_df[tpoint]), aorder)
    row = np.arange(data_df.shape[0])
    col = np.array(data_df[id]) - 1
    val = np.array([1.0] * data_df.shape[0])
    add_mat = csr_matrix((val, (row, col)), shape=(data_df.shape[0], kin_inv.shape[0]))
    zmat_add = []
    for i in range(len(leg_add)):
        zmat_add.append(add_mat.multiply(leg_add[i]))
    
    print('Legendre order for permanent environmental effect:', str(porder))
    leg_per = leg_mt(data_df[tpoint], max(data_df[tpoint]), min(data_df[tpoint]), porder)
    per_mat = csr_matrix((val, (row, col)))
    zmat_per = []
    for i in range(len(leg_per)):
        zmat_per.append((per_mat.multiply(leg_per[i])))
    zmat = [zmat_add, zmat_per]
    kin_inv = [kin_inv, sparse.eye(max(data_df[id]), format="csr")]
    
    return y, xmat, zmat, kin_inv, code_fac_dct[id], id_order, fix_pos_vec



import numpy as np
import pandas as pd
import logging
import os
from glta.gmat import agmat, dgmat_as

logging.basicConfig(level=logging.INFO)

######## 显性基因组关系矩阵
bed_file = '../data/plink'

# 矩阵形式
dgmat0 = dgmat_as(bed_file, inv=True, small_val=0.001, out_fmt='mat')
grm0 = np.load('../data/plink.dgrm0.npz')
giv0 = np.load('../data/plink.dgiv0.npz')
np.savetxt('../data/plink.dgrm0', grm0['mat'])
np.savetxt('../data/plink.dgiv0', giv0['mat'])



# 行-列-值形式
dgmat1 = dgmat_as(bed_file, inv=True, small_val=0.001, out_fmt='row_col_val')
grm1 = np.load('../data/plink.dgrm1.npz')
giv1 = np.load('../data/plink.dgiv1.npz')
grm1_dct = {'row': grm1['row'] + 1,
            'col': grm1['col'] + 1,
            'val': grm1['val']
}
grm1_df = pd.DataFrame(grm1_dct, columns=['row', 'col', 'val'])
grm1_df.to_csv('../data/plink.dgrm1', header=False, index=False, sep=' ')
giv1_dct = {'row': giv1['row'] + 1,
            'col': giv1['col'] + 1,
            'val': giv1['val']
}
giv1_df = pd.DataFrame(giv1_dct, columns=['row', 'col', 'val'])
giv1_df.to_csv('../data/plink.dgiv1', header=False, index=False, sep=' ')



# 个体号-个体号-值形式
dgmat2 = dgmat_as(bed_file, inv=True, small_val=0.001, out_fmt='id_id_val')
grm2 = np.load('../data/plink.dgrm2.npz')
giv2 = np.load('../data/plink.dgiv2.npz')

grm2_dct = {'id0': grm2['id0'],
            'id1': grm2['id1'],
            'val': grm2['val']
}
grm2_df = pd.DataFrame(grm2_dct, columns=['id0', 'id1', 'val'])
grm2_df.to_csv('../data/plink.dgrm2', header=False, index=False, sep=' ')
giv2_dct = {'id0': giv2['id0'],
            'id1': giv2['id1'],
            'val': giv2['val']
}
giv2_df = pd.DataFrame(giv2_dct, columns=['id0', 'id1', 'val'])
giv2_df.to_csv('../data/plink.dgiv2', header=False, index=False, sep=' ')


# 删除
'''
os.remove('../data/plink.dgrm0.npz')
os.remove('../data/plink.dgiv0.npz')
os.remove('../data/plink.dgrm0')
os.remove('../data/plink.dgiv0')

os.remove('../data/plink.dgrm1.npz')
os.remove('../data/plink.dgiv1.npz')
os.remove('../data/plink.dgrm1')
os.remove('../data/plink.dgiv1')

os.remove('../data/plink.dgrm2.npz')
os.remove('../data/plink.dgiv2.npz')
os.remove('../data/plink.dgrm2')
os.remove('../data/plink.dgiv2')
'''



import numpy as np
import pandas as pd
from functools import reduce
import gc
import time
import logging
import sys
from scipy.stats import chi2

from glta.process_plink.process_plink import read_plink, impute_geno


def remma_add_cpu(y, xmat, gmat_lst, var_com, bed_file, out_file='remma_add_cpu'):
    """
    加性检验
    :param y: 表型
    :param xmat: 固定效应设计矩阵
    :param gmat_lst: 基因组关系矩阵列表
    :param var_com: 方差组分
    :param bed_file: plink文件
    :param out_file: 输出文件
    :return:
    """
    logging.info("计算V矩阵及其逆矩阵")
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    vmat_inv = np.linalg.inv(vmat)
    logging.info("计算P矩阵")
    vxmat = np.dot(vmat_inv, xmat)
    xvxmat = np.dot(xmat.T, vxmat)
    xvxmat = np.linalg.inv(xvxmat)
    pmat = reduce(np.dot, [vxmat, xvxmat, vxmat.T])
    pmat = vmat_inv - pmat
    pymat = np.dot(pmat, y)
    pvpmat = reduce(np.dot, [pmat, vmat, pmat])
    del vmat, vmat_inv, pmat
    gc.collect()
    logging.info("读取SNP文件")
    snp_mat = read_plink(bed_file)
    num_id, num_snp = snp_mat.shape
    if np.any(np.isnan(snp_mat)):
        logging.warning('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * num_id)
    freq.shape = (1, num_snp)
    snp_mat = snp_mat - 2 * freq
    scale = np.sum(2*freq*(1-freq))
    logging.info("Scaled factors {:.3f}".format(scale))
    eff_vec = np.dot(snp_mat.T, pymat)[:, -1] * var_com[0]/scale
    var_vec = np.sum(snp_mat * np.dot(pvpmat, snp_mat), axis=0) * var_com[0] * var_com[0]/(scale*scale)
    eff_vec_to_fixed = eff_vec * var_com[0]/(var_vec*scale)  # 转化为固定效应检验计算效应
    chi_vec = eff_vec*eff_vec/var_vec
    p_vec = chi2.sf(chi_vec, 1)
    snp_info_file = bed_file + '.bim'
    snp_info = pd.read_csv(snp_info_file, sep='\s+', header=None)
    res_df = snp_info.iloc[:, [0, 1, 3, 4, 5]]
    res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
    res_df.loc[:, 'eff_val'] = eff_vec
    res_df.loc[:, 'chi_val'] = chi_vec
    res_df.loc[:, 'eff_val_to_fixed'] = eff_vec_to_fixed
    res_df.loc[:, 'p_val'] = p_vec
    try:
        res_df.to_csv(out_file, index=False, header=True, sep=' ')
    except Exception as e:
        logging.error(e)
        sys.exit()
    return res_df



from pandas_plink import read_plink1_bin


def read_plink(bed_file):
    snp_info = read_plink1_bin(bed_file + ".bed", bed_file + ".bim", bed_file + ".fam", verbose=False)
    return snp_info.values


def impute_geno(snp_mat):
    ind = np.where(np.isnan(snp_mat))
    ind_len = len(ind[0])
    snp_mat.val[ind] = np.random.choice([0.0, 1.0, 2.0], ind_len)
    return snp_mat



from setuptools import setup, find_packages

setup(
    cffi_modules=["glta/remma/remma_cpu/_build.py:ffi", "glta/process_plink/_build.py:ffi"],
    name='glta',
    version='2019.12.26',
    description='Genomic Longitudinal Trait Analysis Tool',
    long_description=open('README.rst').read(),
    author='Chao Ning',
    author_email='ningchao91@gmail.com',
    license='BSD License',
    packages=find_packages(),
    platforms=["all"],
    url="https://github.com/chaoning/GLTA",
    include_package_data = True,
    install_requires=[
        'numpy>=1.16.0',
        'pandas>=0.19.0',
        'scipy>=1.1.1',
        'cffi>=1.12.0',
        'pandas_plink>=2.0.0',
    ],

)





import numpy as np
import pandas as pd
from functools import reduce
import gc
import time
from scipy.stats import chi2

from glta.process_plink.process_plink import read_plink, impute_geno


def uvlmm_gwas_add(y, xmat, gmat_lst, var_com, bed_file):
    """
    将SNP效应当成固定效应的单标记检验，可以包含多个关系矩阵（必含加性关系矩阵）
    :param y:
    :param xmat:
    :param gmat_lst:
    :param var_com:
    :param bed_file:
    :return:
    """
    # 计算V矩阵、逆矩阵
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    vmat = np.linalg.inv(vmat)
    # 读取SNP
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        print('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * snp_mat.shape[0])
    freq.shape = (1, snp_mat.shape[1])
    snp_mat = snp_mat - 2 * freq
    # 检验
    eff_vec = []
    chi_vec = []
    p_vec = []
    for i in range(snp_mat.shape[1]):
        xmat_snp = np.concatenate([xmat, snp_mat[:, i:(i+1)]], axis=1)
        vxmat = np.dot(vmat, xmat_snp)
        xvxmat = np.dot(xmat_snp.T, vxmat)
        xvxmat = np.linalg.inv(xvxmat)
        snp_eff = np.dot(xvxmat, np.dot(vxmat.T, y))[-1, -1]
        snp_var = xvxmat[-1, -1]
        snp_chi = snp_eff*snp_eff/snp_var
        p_val = chi2.sf(snp_chi, 1)
        eff_vec.append(snp_eff)
        chi_vec.append(snp_chi)
        p_vec.append(p_val)
    snp_info_file = bed_file + '.bim'
    snp_info = pd.read_csv(snp_info_file, sep='\s+', header=None)
    res_df = snp_info.iloc[:, [0, 1, 3, 4, 5]]
    res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
    res_df.loc[:, 'eff_val'] = eff_vec
    res_df.loc[:, 'chi_val'] = chi_vec
    res_df.loc[:, 'p_val'] = p_vec
    return res_df


def uvlmm_gwas_add_eigen(y, xmat, agmat, var_com, bed_file):
    """
    将SNP效应当成固定效应的单标记检验，仅包含加性基因组关系矩阵，特征分解加快运算速度
    :param y:
    :param xmat:
    :param agmat:
    :param var_com:
    :param bed_file:
    :return:
    """


def uvlmm_gwas_dom(y, xmat, gmat_lst, var_com, bed_file):
    """
    显性检验，可以包含多个关系矩阵（必含加性关系矩阵）
    :param y:
    :param xmat:
    :param gmat_lst:
    :param var_com:
    :param bed_file:
    :return:
     """
    # 计算V矩阵、逆矩阵
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    vmat = np.linalg.inv(vmat)
    # 读取SNP
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        print('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * snp_mat.shape[0])
    freq.shape = (1, snp_mat.shape[1])
    snp_mat_add = snp_mat - 2 * freq
    snp_mat[snp_mat > 1.5] = 0.0
    snp_mat = snp_mat - 2 * freq * (1 - freq)
    # 检验
    eff_vec = []
    chi_vec = []
    p_vec = []
    for i in range(snp_mat.shape[1]):
        xmat_snp = np.concatenate([xmat, snp_mat_add[:, i:(i+1)], snp_mat[:, i:(i+1)]], axis=1)
        vxmat = np.dot(vmat, xmat_snp)
        xvxmat = np.dot(xmat_snp.T, vxmat)
        xvxmat = np.linalg.inv(xvxmat)
        snp_eff = np.dot(xvxmat, np.dot(vxmat.T, y))[-1, -1]
        snp_var = xvxmat[-1, -1]
        snp_chi = snp_eff*snp_eff/snp_var
        p_val = chi2.sf(snp_chi, 1)
        eff_vec.append(snp_eff)
        chi_vec.append(snp_chi)
        p_vec.append(p_val)
    snp_info_file = bed_file + '.bim'
    snp_info = pd.read_csv(snp_info_file, sep='\s+', header=None)
    res_df = snp_info.iloc[:, [0, 1, 3, 4, 5]]
    res_df.columns = ['chro', 'snp_ID', 'pos', 'allele1', 'allele2']
    res_df.loc[:, 'eff_val'] = eff_vec
    res_df.loc[:, 'chi_val'] = chi_vec
    res_df.loc[:, 'p_val'] = p_vec
    return res_df

def uvlmm_gwas_dom_eigen(y, xmat, agmat, var_com, bed_file):
    """
    显性检验，仅包含加性基因组关系矩阵，特征分解加快运算速度
    :param y:
    :param xmat:
    :param agmat:
    :param var_com:
    :param bed_file:
    :return:
    """


def uvlmm_gwas_epiAA(y, xmat, gmat_lst, var_com, bed_file):
    """
    加加上位，将SNP效应当成固定效应的单标记检验，可以包含多个关系矩阵（必含加性关系矩阵）
    :param y:
    :param xmat:
    :param gmat_lst:
    :param var_com:
    :param bed_file:
    :return:
    """
    # 计算V矩阵、逆矩阵
    y = np.array(y).reshape(-1, 1)
    n = y.shape[0]
    xmat = np.array(xmat).reshape(n, -1)
    vmat = np.diag([var_com[-1]] * n)
    for val in range(len(gmat_lst)):
        vmat += gmat_lst[val] * var_com[val]
    vmat = np.linalg.inv(vmat)
    # 读取SNP
    snp_mat = read_plink(bed_file)
    if np.any(np.isnan(snp_mat)):
        print('Missing genotypes are imputed with random genotypes.')
        snp_mat = impute_geno(snp_mat)
    freq = np.sum(snp_mat, axis=0) / (2 * snp_mat.shape[0])
    freq.shape = (1, snp_mat.shape[1])
    snp_mat = snp_mat - 2 * freq
    # 检验
    snpi = []
    snpj = []
    eff_vec = []
    chi_vec = []
    p_vec = []
    for i in range(snp_mat.shape[1]-1):
        for j in range(i+1, snp_mat.shape[1]):
            snpi.append(i)
            snpj.append(j)
            xmat_snp = np.concatenate([xmat, snp_mat[:, i:(i + 1)], snp_mat[:, j:(j + 1)], snp_mat[:, i:(i + 1)]*snp_mat[:, j:(j + 1)]], axis=1)
            vxmat = np.dot(vmat, xmat_snp)
            xvxmat = np.dot(xmat_snp.T, vxmat)
            xvxmat = np.linalg.inv(xvxmat)
            snp_eff = np.dot(xvxmat, np.dot(vxmat.T, y))[-1, -1]
            snp_var = xvxmat[-1, -1]
            snp_chi = snp_eff * snp_eff / snp_var
            p_val = chi2.sf(snp_chi, 1)
            eff_vec.append(snp_eff)
            chi_vec.append(snp_chi)
            p_vec.append(p_val)
    res_df = pd.DataFrame({
        'snpi': snpi,
        'snpj': snpj,
        'snp_eff': eff_vec,
        'p_val': p_vec
    })
    return res_df
